@@ -316,9 +316,6 @@ return 1
     .. _Lua: http://www.lua.org
     """
 
-    #: This driver requires constant periodic (heart) beatings.
-    requires_beating = True
-
     def __init__(self, member_id, parsed_url, options):
         super(RedisDriver, self).__init__()
         options = utils.collapse(options, exclude=self.CLIENT_LIST_ARGS)
@@ -511,6 +508,7 @@ return 1
             except coordination.ToozError:
                 LOG.warning("Unable to heartbeat lock '%s'", lock,
                             exc_info=True)
+        return min(self.lock_timeout, self.membership_timeout)
 
     def _stop(self):
         while self._acquired_locks:
