@@ -22,6 +22,7 @@ from kazoo import exceptions
 from kazoo.handlers import eventlet as eventlet_handler
 from kazoo.handlers import threading as threading_handler
 from kazoo.protocol import paths
+from oslo_utils import encodeutils
 from oslo_utils import strutils
 import six
 from six.moves import filter as compat_filter
@@ -36,7 +37,6 @@ class ZooKeeperLock(locking.Lock):
         super(ZooKeeperLock, self).__init__(name)
         self._lock = lock
         self._client = lock.client
-        self.acquired = False
 
     def is_still_owner(self):
         if not self.acquired:
@@ -57,18 +57,20 @@ class ZooKeeperLock(locking.Lock):
 
     def acquire(self, blocking=True):
         blocking, timeout = utils.convert_blocking(blocking)
-        self.acquired = self._lock.acquire(blocking=blocking,
-                                           timeout=timeout)
-        return self.acquired
+        return self._lock.acquire(blocking=blocking,
+                                  timeout=timeout)
 
     def release(self):
         if self.acquired:
             self._lock.release()
-            self.acquired = False
             return True
         else:
             return False
 
+    @property
+    def acquired(self):
+        return self._lock.is_acquired
+
 
 class BaseZooKeeperDriver(coordination.CoordinationDriver):
     """Initialize the zookeeper driver.
@@ -91,17 +93,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             self._coord.start(timeout=self.timeout)
         except self._coord.handler.timeout_exception as e:
+            e_msg = encodeutils.exception_to_unicode(e)
             coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          "operation error: %s" % (e),
+                                          "Operational error: %s" % e_msg,
                                           cause=e)
-
         try:
-            self._coord.ensure_path(self.paths_join("/", self._namespace))
+            self._coord.ensure_path(self._paths_join("/", self._namespace))
         except exceptions.KazooException as e:
+            e_msg = encodeutils.exception_to_unicode(e)
             coordination.raise_with_cause(coordination.ToozError,
-                                          "operation error: %s" % (e),
+                                          "Operational error: %s" % e_msg,
                                           cause=e)
-
         self._group_members = collections.defaultdict(set)
         self._watchers = collections.deque()
         self._leader_locks = {}
@@ -117,22 +119,24 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
     def _loads(blob):
         return utils.loads(blob)
 
-    @staticmethod
-    def _create_group_handler(async_result, timeout,
+    def _create_group_handler(self, async_result, timeout,
                               timeout_exception, group_id):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NodeExistsError:
             raise coordination.GroupAlreadyExist(group_id)
-        except exceptions.NoNodeError:
-            raise coordination.ToozError("tooz namespace has not been created")
+        except exceptions.NoNodeError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "Tooz namespace '%s' has not"
+                                          " been created" % self._namespace,
+                                          cause=e)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
     def create_group(self, group_id):
@@ -149,7 +153,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
@@ -157,7 +161,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             raise coordination.GroupNotEmpty(group_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
     def delete_group(self, group_id):
@@ -174,7 +178,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NodeExistsError:
             raise coordination.MemberAlreadyExist(group_id, member_id)
@@ -182,7 +186,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
     def join_group(self, group_id, capabilities=b""):
@@ -202,15 +206,34 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
+    def heartbeat(self):
+        # Just fetch the base path (and do nothing with it); this will
+        # force any waiting heartbeat responses to be flushed, and also
+        # ensures that the connection still works as expected...
+        base_path = self._paths_join("/", self._namespace)
+        try:
+            self._coord.get(base_path)
+        except self._timeout_exception as e:
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        except exceptions.NoNodeError:
+            pass
+        except exceptions.ZookeeperError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        return self.timeout
+
     def leave_group(self, group_id):
         member_path = self._path_member(group_id, self._member_id)
         async_result = self._coord.delete_async(member_path)
@@ -225,19 +248,19 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             members_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         else:
             return set(m.encode('ascii') for m in members_ids)
 
     def get_members(self, group_id):
-        group_path = self.paths_join("/", self._namespace, group_id)
+        group_path = self._paths_join("/", self._namespace, group_id)
         async_result = self._coord.get_children_async(group_path)
         return ZooAsyncResult(async_result, self._get_members_handler,
                               timeout_exception=self._timeout_exception,
@@ -250,13 +273,13 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
     def update_capabilities(self, group_id, capabilities):
@@ -275,13 +298,13 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             capabilities = async_result.get(block=True, timeout=timeout)[0]
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         else:
             return cls._loads(capabilities)
@@ -303,13 +326,13 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                                                          timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         else:
             member_info = {
@@ -327,45 +350,49 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                               timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
-    @staticmethod
-    def _get_groups_handler(async_result, timeout, timeout_exception):
+    def _get_groups_handler(self, async_result, timeout, timeout_exception):
         try:
             group_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
             coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        except exceptions.NoNodeError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "Tooz namespace '%s' has not"
+                                          " been created" % self._namespace,
                                           cause=e)
-        except exceptions.NoNodeError:
-            raise coordination.ToozError("tooz namespace has not been created")
         except exceptions.ZookeeperError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         else:
             return set(g.encode('ascii') for g in group_ids)
 
     def get_groups(self):
-        tooz_namespace = self.paths_join("/", self._namespace)
+        tooz_namespace = self._paths_join("/", self._namespace)
         async_result = self._coord.get_children_async(tooz_namespace)
         return ZooAsyncResult(async_result, self._get_groups_handler,
                               timeout_exception=self._timeout_exception)
 
     def _path_group(self, group_id):
-        return self.paths_join("/", self._namespace, group_id)
+        return self._paths_join("/", self._namespace, group_id)
 
     def _path_member(self, group_id, member_id):
-        return self.paths_join("/", self._namespace,
-                               group_id, member_id)
+        return self._paths_join("/", self._namespace, group_id, member_id)
 
     @staticmethod
-    def paths_join(*args):
-        lpaths = []
+    def _paths_join(arg, *more_args):
+        """Converts paths into a string (unicode)."""
+        args = [arg]
+        args.extend(more_args)
+        cleaned_args = []
         for arg in args:
             if isinstance(arg, six.binary_type):
-                lpaths.append(arg.decode('ascii'))
+                cleaned_args.append(arg.decode('ascii'))
             else:
-                lpaths.append(arg)
-        return paths.join(*lpaths)
+                cleaned_args.append(arg)
+        return paths.join(*cleaned_args)
 
 
 class KazooDriver(BaseZooKeeperDriver):
@@ -407,6 +434,20 @@ class KazooDriver(BaseZooKeeperDriver):
     this option is one of the keys in this dictionary).
     """
 
+    CHARACTERISTICS = (
+        coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
+        # Writes *always* go through a single leader process, but it may
+        # take a while for those writes to propagate to followers (and =
+        # during this time clients can read older values)...
+        coordination.Characteristics.SEQUENTIAL,
+    )
+    """
+    Tuple of :py:class:`~tooz.coordination.Characteristics` introspectable
+    enum member(s) that can be used to interogate how this driver works.
+    """
+
     def __init__(self, member_id, parsed_url, options):
         super(KazooDriver, self).__init__(member_id, parsed_url, options)
         self._coord = self._make_client(parsed_url, self._options)
@@ -557,7 +598,7 @@ class KazooDriver(BaseZooKeeperDriver):
 
     def get_lock(self, name):
         z_lock = self._coord.Lock(
-            self.paths_join(b"/", self._namespace, b"locks", name),
+            self._paths_join(b"/", self._namespace, b"locks", name),
             self._member_id.decode('ascii'))
         return ZooKeeperLock(name, z_lock)
 
