@@ -139,7 +139,7 @@ class TestAPI(testscenarios.TestWithScenarios,
             self._coord.create_group(group_id).get()
         created_groups = self._coord.get_groups().get()
         for group_id in groups_ids:
-            self.assertTrue(group_id in created_groups)
+            self.assertIn(group_id, created_groups)
 
     def test_delete_group(self):
         self._coord.create_group(self.group_id).get()
@@ -338,8 +338,28 @@ class TestAPI(testscenarios.TestWithScenarios,
                              update_cap.get)
 
     def test_heartbeat(self):
+        if not self._coord.requires_beating:
+            raise testcase.TestSkipped("Test not applicable (heartbeating"
+                                       " not required)")
         self._coord.heartbeat()
 
+    def test_heartbeat_loop(self):
+        if not self._coord.requires_beating:
+            raise testcase.TestSkipped("Test not applicable (heartbeating"
+                                       " not required)")
+
+        heart = self._coord.heart
+        self.assertFalse(heart.is_alive())
+        heart.start()
+
+        # This will timeout if nothing ever is done...
+        try:
+            while not heart.beats:
+                time.sleep(1)
+        finally:
+            heart.stop()
+            heart.wait()
+
     def test_disconnect_leave_group(self):
         member_id_test2 = self._get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
